@@ -1,11 +1,16 @@
-import click
-import questionary
-from pathlib import Path
-import requests
 import subprocess
 import os
 import neuronum
 import json
+import platform
+import glob
+import asyncio
+import aiohttp
+import click
+import questionary
+from pathlib import Path
+import requests
+
 
 @click.group()
 def cli():
@@ -81,7 +86,6 @@ def create_cell():
             click.echo(f"Error:'{email}' already assigned!")
 
 
-
 @click.command()
 def connect_cell():
     email = click.prompt("Enter your Email")
@@ -240,25 +244,19 @@ def delete_cell():
         click.echo(f"Neuronum Cell '{host}' has been deleted!")
     else: 
         click.echo(f"Neuronum Cell '{host}' deletion failed!")
- 
+
 
 @click.command()
-@click.option('--sync', default=None, help="Optional stream ID to generate the sync template.")
-def init_node(sync):
-
-    node_type = questionary.select(
-        "Choose Node type:",
-        choices=["public", "private"]
-    ).ask()
-
-    descr = click.prompt("Node description (max. 25 characters)")
-
-    stream = sync if sync else "n9gW3LxQcecI::stx"
+@click.option('--sync', multiple=True, default=None, help="Optional stream IDs for sync.")
+@click.option('--stream', multiple=True, default=None, help="Optional stream ID for stream.")
+def init_node(sync, stream):
+    asyncio.run(async_init_node(sync, stream))
 
+async def async_init_node(sync, stream):
     credentials_folder_path = Path.home() / ".neuronum"
     env_path = credentials_folder_path / ".env"
 
-    env_data = {}
+    env_data = {}  
 
     try:
         with open(env_path, "r") as f:
@@ -278,63 +276,61 @@ def init_node(sync):
         click.echo(f"Error reading .env file: {e}")
         return
 
-    cell = neuronum.Cell(
-    host=host,         
-    password=password,                         
-    network=network,                        
-    synapse=synapse
-    )
+    url = f"https://{network}/api/init_node"
+    node_payload = {"host": host, "password": password, "synapse": synapse}
 
-    tx = cell.list_tx()
-    ctx = cell.list_ctx()
-    stx = cell.list_stx()
-    contracts = cell.list_contracts()
-    nodes = cell.list_nodes()
-
-    url = f"https://{network}/api/init_node/{node_type}"
-
-    node = {"descr": descr, "host": host, "password": password, "synapse": synapse}
-
-    try:
-        response = requests.post(url, json=node)
-        response.raise_for_status()
-        nodeID = response.json()["nodeID"]
-    except requests.exceptions.RequestException as e:
-        click.echo(f"Error sending request: {e}")
-        return
+    async with aiohttp.ClientSession() as session:
+        try:
+            async with session.post(url, json=node_payload) as response:
+                response.raise_for_status()
+                data = await response.json()
+                nodeID = data["nodeID"]
+        except aiohttp.ClientError as e:
+            click.echo(f"Error sending request: {e}")
+            return
 
     node_filename = "node_" + nodeID.replace("::node", "")
     project_path = Path(node_filename)
     project_path.mkdir(exist_ok=True)
 
+    cell = neuronum.Cell(
+        host=host,         
+        password=password,                         
+        network=network,                        
+        synapse=synapse
+    )
+
+    cells = await cell.list_cells()
+    tx = await cell.list_tx()
+    ctx = await cell.list_ctx()
+    stx = await cell.list_stx()
+    contracts = await cell.list_contracts()
+    nodes = await cell.list_nodes()
+
+    await asyncio.to_thread((project_path / "cells.json").write_text, json.dumps(cells, indent=4))
+    await asyncio.to_thread((project_path / "transmitters.json").write_text, json.dumps(tx, indent=4))
+    await asyncio.to_thread((project_path / "circuits.json").write_text, json.dumps(ctx, indent=4))
+    await asyncio.to_thread((project_path / "streams.json").write_text, json.dumps(stx, indent=4))
+    await asyncio.to_thread((project_path / "contracts.json").write_text, json.dumps(contracts, indent=4))
+    await asyncio.to_thread((project_path / "nodes.json").write_text, json.dumps(nodes, indent=4))
+
     env_path = project_path / ".env"
-    env_path.write_text(f"NODE={nodeID}\nHOST={host}\nPASSWORD={password}\nNETWORK={network}\nSYNAPSE={synapse}\n")
+    await asyncio.to_thread(env_path.write_text, f"NODE={nodeID}\nHOST={host}\nPASSWORD={password}\nNETWORK={network}\nSYNAPSE={synapse}\n")
 
     gitignore_path = project_path / ".gitignore"
-    gitignore_path.write_text(f".env\n")
-
-    tx_path = project_path / "transmitters.json"
-    tx_path.write_text(json.dumps(tx, indent=4))
-
-    ctx_path = project_path / "circuits.json"
-    ctx_path.write_text(json.dumps(ctx, indent=4))
-
-    stx_path = project_path / "streams.json"
-    stx_path.write_text(json.dumps(stx, indent=4))
-
-    contracts_path = project_path / "contracts.json"
-    contracts_path.write_text(json.dumps(contracts, indent=4))
-
-    nodes_path = project_path / "nodes.json"
-    nodes_path.write_text(json.dumps(nodes, indent=4))
+    await asyncio.to_thread(gitignore_path.write_text, ".env\n")
 
     nodemd_path = project_path / "NODE.md"
-    nodemd_path.write_text("""\
-## Use this NODE.md file to add instructions on how to interact with your node                        
-""")
-        
-    main_path = project_path / "main.py"
-    main_path.write_text(f"""\
+    await asyncio.to_thread(nodemd_path.write_text, "## Use this NODE.md file to add instructions on how to interact with your node\n")
+
+
+    stx = sync[0] if sync else (stream[0] if stream else host.replace("::cell", "::stx"))
+
+    if sync:
+        for stx in sync:
+            sync_path = project_path / f"sync_{stx.replace('::stx', '')}.py"
+            sync_path.write_text(f"""\
+import asyncio
 import neuronum
 import os
 from dotenv import load_dotenv
@@ -352,15 +348,145 @@ cell = neuronum.Cell(
     synapse=synapse
 )
 
-STX = "{stream}"
-stream = cell.sync(STX)
-for operation in stream:
-    label = operation.get("label")
-    value = operation.get("data").get("message")
-    ts = operation.get("time")
-    stxID = operation.get("stxID")
-    operator = operation.get("operator")
-    print(label, value, ts, stxID, operator)                              
+async def main():
+    STX = "{stx}"
+    async for operation in cell.sync(STX):
+        label = operation.get("label")
+        data = operation.get("data")
+        ts = operation.get("time")
+        stxID = operation.get("stxID")
+        operator = operation.get("operator")
+        print(label, data, ts, stxID, operator)
+
+asyncio.run(main())
+""")
+
+
+    if stream:
+        for stx in stream:
+            stream_path = project_path / f"stream_{stx.replace('::stx', '')}.py"
+            stream_path.write_text(f"""\
+import asyncio
+import neuronum
+import os
+from dotenv import load_dotenv
+
+load_dotenv()
+host = os.getenv("HOST")
+password = os.getenv("PASSWORD")
+network = os.getenv("NETWORK")
+synapse = os.getenv("SYNAPSE")
+
+cell = neuronum.Cell(
+    host=host,
+    password=password,
+    network=network,
+    synapse=synapse
+)
+
+async def main():
+    STX = "{stx}"
+    label = "your_label"
+    
+    while True:
+        data = {{
+            "key1": "value1",
+            "key2": "value2",
+            "key3": "value3",
+        }}
+        await cell.stream(label, data, STX)
+
+asyncio.run(main())
+""")
+    
+    if not sync and not stream:
+        sync_path = project_path / f"sync_{stx.replace('::stx', '')}.py"
+        sync_path.write_text(f"""\
+import asyncio
+import neuronum
+import os
+from dotenv import load_dotenv
+
+load_dotenv()
+host = os.getenv("HOST")
+password = os.getenv("PASSWORD")
+network = os.getenv("NETWORK")
+synapse = os.getenv("SYNAPSE")
+
+cell = neuronum.Cell(
+    host=host,
+    password=password,
+    network=network,
+    synapse=synapse
+)
+
+async def main():
+    STX = "{stx}"
+    async for operation in cell.sync(STX):
+        message = operation.get("data").get("message")
+        print(message)
+
+asyncio.run(main())
+""")
+        
+        stream_path = project_path / f"stream_{stx.replace('::stx', '')}.py"
+        stream_path.write_text(f"""\
+import asyncio
+import neuronum
+import os
+from dotenv import load_dotenv
+
+load_dotenv()
+host = os.getenv("HOST")
+password = os.getenv("PASSWORD")
+network = os.getenv("NETWORK")
+synapse = os.getenv("SYNAPSE")
+
+cell = neuronum.Cell(
+    host=host,
+    password=password,
+    network=network,
+    synapse=synapse
+)
+
+async def main():
+    STX = "{stx}"
+    label = "Welcome to Neuronum"
+    
+    while True:
+        data = {{
+            "message": "Hello, Neuronum!"
+        }}
+        await cell.stream(label, data, STX)
+
+asyncio.run(main())
+""")
+        
+    scan_path = project_path / f"scan.py"
+    scan_path.write_text(f"""\
+import asyncio
+import neuronum
+import os
+from dotenv import load_dotenv
+
+load_dotenv()
+host = os.getenv("HOST")
+password = os.getenv("PASSWORD")
+network = os.getenv("NETWORK")
+synapse = os.getenv("SYNAPSE")
+
+cell = neuronum.Cell(
+    host=host,
+    password=password,
+    network=network,
+    synapse=synapse
+)
+
+async def main():
+    async for cp in cell.scan():
+        print(cp)
+
+asyncio.run(main())
 """)
 
     click.echo(f"Neuronum Node '{nodeID}' initialized!")
@@ -368,39 +494,76 @@ for operation in stream:
 
 @click.command()
 def start_node():
+    scan_type = questionary.select(
+        "Scan for Neuronum Cells and Nodes (Ensure Bluetooth is enabled)",
+        choices=["On", "Off"]
+    ).ask()
+
     click.echo("Starting Node...")
 
     project_path = Path.cwd()
-    main_file = project_path / "main.py"
+    script_files = glob.glob("sync_*.py") + glob.glob("stream_*.py")
 
-    if not main_file.exists():
-        click.echo("Error: main.py not found. Make sure the node is set up.")
+    if scan_type == "On":
+        script_files += glob.glob("scan.py")
+
+    processes = []
+
+    for script in script_files:
+        script_path = project_path / script
+        if script_path.exists():
+            process = subprocess.Popen(["python", str(script_path)], start_new_session=True)
+            processes.append(process.pid)
+
+    if not processes:
+        click.echo("Error: No valid node script found. Ensure the node is set up correctly.")
         return
 
-    process = subprocess.Popen(["python", str(main_file)], start_new_session=True)
-
     with open("node_pid.txt", "w") as f:
-        f.write(str(process.pid))
+        f.write("\n".join(map(str, processes)))
 
     click.echo("Node started successfully!")
 
 
+
 @click.command()
 def stop_node():
+    asyncio.run(async_stop_node())
+
+async def async_stop_node():
     click.echo("Stopping Node...")
 
+    node_pid_path = Path("node_pid.txt")
+
     try:
         with open("node_pid.txt", "r") as f:
-            pid = int(f.read().strip())
-        os.kill(pid, 9) 
-        os.remove("node_pid.txt")
+            pids = [int(pid.strip()) for pid in f.readlines()]
+
+        system_name = platform.system()
+
+        for pid in pids:
+            try:
+                if system_name == "Windows":
+                    await asyncio.to_thread(subprocess.run, ["taskkill", "/F", "/PID", str(pid)], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
+                else:
+                    await asyncio.to_thread(os.kill, pid, 9)
+            except ProcessLookupError:
+                click.echo(f"Warning: Process {pid} already stopped or does not exist.")
+
+        await asyncio.to_thread(os.remove, node_pid_path)
         click.echo("Node stopped successfully!")
+
     except FileNotFoundError:
         click.echo("Error: No active node process found.")
+    except subprocess.CalledProcessError:
+        click.echo("Error: Unable to stop some node processes.")
 
 
 @click.command()
 def register_node():
+    asyncio.run(async_register_node())
+
+async def async_register_node():
     env_data = {}
     try:
         with open(".env", "r") as f:
@@ -420,6 +583,13 @@ def register_node():
     except Exception as e:
         print(f"Error reading .env file: {e}")
         return
+    
+    node_type = questionary.select(
+        "Choose Node type:",
+        choices=["public", "private"]
+    ).ask()
+
+    descr = click.prompt("Node description (max. 25 characters)")
 
     try:
         with open("NODE.md", "r") as f: 
@@ -432,31 +602,38 @@ def register_node():
         print(f"Error reading NODE.md file: {e}")
         return
 
-    url = f"https://{network}/api/register_node"
+    url = f"https://{network}/api/register_node/{node_type}"
 
     node = {
         "nodeID": nodeID,
+        "descr": descr,
         "host": host,
         "password": password,
         "synapse": synapse,
         "nodemd_file": nodemd_file
     }
 
-    try:
-        response = requests.post(url, json=node)
-        response.raise_for_status()
-        nodeID = response.json()["nodeID"]
-        node_url = response.json()["node_url"]
-    except requests.exceptions.RequestException as e:
-        click.echo(f"Error sending request: {e}")
-        return
+    async with aiohttp.ClientSession() as session:
+        try:
+            async with session.post(url, json=node) as response:
+                response.raise_for_status()
+                data = await response.json()
+                nodeID = data["nodeID"]
+                node_url = data["node_url"]
+        except aiohttp.ClientError as e:
+            click.echo(f"Error sending request: {e}")
+            return
 
     click.echo(f"Neuronum Node '{nodeID}' registered! Visit: {node_url}")
 
 
 @click.command()
 def update_node():
+    asyncio.run(async_update_node())
+
+async def async_update_node():
     env_data = {}
+
     try:
         with open(".env", "r") as f:
             for line in f:
@@ -470,26 +647,25 @@ def update_node():
         synapse = env_data.get("SYNAPSE", "")
 
     except FileNotFoundError:
-        print("Error: .env with credentials not found")
+        click.echo("Error: .env with credentials not found")
         return
     except Exception as e:
-        print(f"Error reading .env file: {e}")
+        click.echo(f"Error reading .env file: {e}")
         return
 
     try:
-        with open("NODE.md", "r") as f: 
-            nodemd_file = f.read() 
+        with open("NODE.md", "r") as f:
+            nodemd_file = f.read()
 
     except FileNotFoundError:
-        print("Error: NODE.md file not found")
+        click.echo("Error: NODE.md file not found")
         return
     except Exception as e:
-        print(f"Error reading NODE.md file: {e}")
+        click.echo(f"Error reading NODE.md file: {e}")
         return
 
     url = f"https://{network}/api/update_node"
-
-    node = {
+    node_payload = {
         "nodeID": nodeID,
         "host": host,
         "password": password,
@@ -497,47 +673,48 @@ def update_node():
         "nodemd_file": nodemd_file
     }
 
-    try:
-        response = requests.post(url, json=node)
-        response.raise_for_status()
-        nodeID = response.json()["nodeID"]
-        node_url = response.json()["node_url"]
-    except requests.exceptions.RequestException as e:
-        click.echo(f"Error sending request: {e}")
-        return
-    
+    async with aiohttp.ClientSession() as session:
+        try:
+            async with session.post(url, json=node_payload) as response:
+                response.raise_for_status()
+                data = await response.json()
+                nodeID = data["nodeID"]
+                node_url = data["node_url"]
+        except aiohttp.ClientError as e:
+            click.echo(f"Error sending request: {e}")
+            return
 
     cell = neuronum.Cell(
-    host=host,         
-    password=password,                         
-    network=network,                        
-    synapse=synapse
+        host=host,
+        password=password,
+        network=network,
+        synapse=synapse
     )
 
-    tx = cell.list_tx()
-    ctx = cell.list_ctx()
-    stx = cell.list_stx()
-    contracts = cell.list_contracts()
-    nodes = cell.list_nodes()
+    cells = await cell.list_cells()
+    tx = await cell.list_tx()
+    ctx = await cell.list_ctx()
+    stx = await cell.list_stx()
+    contracts = await cell.list_contracts()
+    nodes = await cell.list_nodes()
 
-    tx_path = Path("transmitters.json")
-    ctx_path = Path("circuits.json")
-    stx_path = Path("streams.json")
-    contracts_path = Path("contracts.json")
-    nodes_path = Path("nodes.json")
-
-    tx_path.write_text(json.dumps(tx, indent=4))
-    ctx_path.write_text(json.dumps(ctx, indent=4))
-    stx_path.write_text(json.dumps(stx, indent=4))
-    contracts_path.write_text(json.dumps(contracts, indent=4))
-    nodes_path.write_text(json.dumps(nodes, indent=4))
+    await asyncio.to_thread(Path("cells.json").write_text, json.dumps(cells, indent=4))
+    await asyncio.to_thread(Path("transmitters.json").write_text, json.dumps(tx, indent=4))
+    await asyncio.to_thread(Path("circuits.json").write_text, json.dumps(ctx, indent=4))
+    await asyncio.to_thread(Path("streams.json").write_text, json.dumps(stx, indent=4))
+    await asyncio.to_thread(Path("contracts.json").write_text, json.dumps(contracts, indent=4))
+    await asyncio.to_thread(Path("nodes.json").write_text, json.dumps(nodes, indent=4))
 
     click.echo(f"Neuronum Node '{nodeID}' updated! Visit: {node_url}")
 
 
 @click.command()
 def delete_node():
+    asyncio.run(async_delete_node())
+
+async def async_delete_node():
     env_data = {}
+
     try:
         with open(".env", "r") as f:
             for line in f:
@@ -551,32 +728,42 @@ def delete_node():
         synapse = env_data.get("SYNAPSE", "")
 
     except FileNotFoundError:
-        print("Error: .env with credentials not found")
+        click.echo("Error: .env with credentials not found")
         return
     except Exception as e:
-        print(f"Error reading .env file: {e}")
+        click.echo(f"Error reading .env file: {e}")
         return
 
     url = f"https://{network}/api/delete_node"
-
-    node = {
+    node_payload = {
         "nodeID": nodeID,
         "host": host,
         "password": password,
         "synapse": synapse
     }
 
-    try:
-        response = requests.post(url, json=node)
-        response.raise_for_status()
-        nodeID = response.json()["nodeID"]
-    except requests.exceptions.RequestException as e:
-        click.echo(f"Error sending request: {e}")
-        return
+    async with aiohttp.ClientSession() as session:
+        try:
+            async with session.post(url, json=node_payload) as response:
+                response.raise_for_status()
+                data = await response.json()
+                nodeID = data["nodeID"]
+        except aiohttp.ClientError as e:
+            click.echo(f"Error sending request: {e}")
+            return
 
     click.echo(f"Neuronum Node '{nodeID}' deleted!")
 
 
+@click.command() 
+def call_cellai(): 
+    try: 
+        from cellai import cellai 
+        cellai.main() 
+    except ImportError: 
+        click.echo("Cellai not found. Please check the necessary dependencies.")
+
+
 cli.add_command(create_cell)
 cli.add_command(connect_cell)
 cli.add_command(view_cell)
@@ -588,6 +775,7 @@ cli.add_command(stop_node)
 cli.add_command(register_node)
 cli.add_command(update_node)
 cli.add_command(delete_node)
+cli.add_command(call_cellai)
 
 
 if __name__ == "__main__":
